"""Test to make sure the envelope function works."""

import pytest
import numpy as np
from iotile_analytics.core.utilities import envelope


def test_basic_envelope():
    """Make sure envelope works correctly."""

    x_vals = np.linspace(0, 10, 10)

    in1 = np.ndarray([10, 2])
    in1[:, 0] = x_vals
    in1[:5, 1] = 1.0
    in1[5:, 1] = -1.0

    in2 = np.ndarray([10, 2])
    in2[:, 0] = x_vals
    in2[:5, 1] = -1.0
    in2[5:, 1] = 1.0

    out = envelope(in1, in2, bin_count=10)
    assert out.shape[0] == 10
    assert out.shape[1] == 3

    assert out[:, 1] == pytest.approx(-1.0*np.ones_like(out[:, 1]))
    assert out[:, 2] == pytest.approx(np.ones_like(out[:, 1]))


def test_envelope_interpolation():
    """Make sure we can fill in gaps in the envelope."""

    x_vals = np.linspace(0, 10, 10)

    in1 = np.ndarray([10, 2])
    in1[:, 0] = x_vals
    in1[:5, 1] = 1.0
    in1[5:, 1] = -1.0

    in2 = np.ndarray([10, 2])
    in2[:, 0] = x_vals
    in2[:5, 1] = -1.0
    in2[5:, 1] = 1.0

    out = envelope(in1, in2, bin_count=100)
    assert out.shape[0] == 100
    assert out.shape[1] == 3

    assert out[:, 1] == pytest.approx(-1.0*np.ones_like(out[:, 1]))
    assert out[:, 2] == pytest.approx(np.ones_like(out[:, 1]))


def test_bin_edges():
    """Make sure we return the right bin edges in all 3 supported cases."""

    x_vals = np.linspace(0, 10, 10)

    in1 = np.ndarray([10, 2])
    in1[:, 0] = x_vals
    in1[:5, 1] = 1.0
    in1[5:, 1] = -1.0

    in2 = np.ndarray([10, 2])
    in2[:, 0] = x_vals
    in2[:5, 1] = -1.0
    in2[5:, 1] = 1.0

    out_left = envelope(in1, in2, bin_count=10, bin_mark="left")
    out_right = envelope(in1, in2, bin_count=10, bin_mark="right")
    out_center = envelope(in1, in2, bin_count=10, bin_mark="center")

    bins = bins = np.linspace(0, 10, 11)

    assert out_left[:, 0] == pytest.approx(bins[:-1])
    assert out_right[:, 0] == pytest.approx(bins[1:])
    assert out_center[:, 0] == pytest.approx((bins[1:] + bins[:-1]) / 2.0)



"""Make sure the analysis group works."""

from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

import pytest
from typedargs.exceptions import ArgumentError
from iotile_analytics.core import CloudSession, AnalysisGroup
from iotile_analytics.core.exceptions import CloudError, AuthenticationError


def test_session_login(water_meter):
    """Make sure we can login successfully."""

    domain, _cloud = water_meter

    session = CloudSession('test@arch-iot.com', 'test', domain=domain)
    assert session.token == "JWT_USER"

    with pytest.raises(AuthenticationError):
        CloudSession('test@arch-iot.com', 'test2', domain=domain)


def test_device_analysis(water_meter):
    """Make sure we can create an analysis group from a device.

    Also tests to make sure we properly read in stream time series and event
    data.  We support reading stream data from json API dumps or from CSV
    files.
    """

    domain, _cloud = water_meter
    CloudSession('test@arch-iot.com', 'test', domain=domain)

    group = AnalysisGroup.FromDevice('d--0000-0000-0000-00d2', domain=domain)

    assert group.stream_counts['s--0000-0077--0000-0000-0000-00d2--5001']['points'] == 11
    assert group.stream_counts['s--0000-0077--0000-0000-0000-00d2--5002']['points'] == 3


def test_stream_download(filter_group):
    """Make sure we can download streams from an analysis group."""

    data = filter_group.fetch_stream('5001')
    assert len(data) == 11


def test_invalid_stream(filter_group):
    """Make sure we raise the right error if we can't find a stream."""

    with pytest.raises(ArgumentError):
        filter_group.fetch_stream('5500')



from ..environment import Environment
from typedargs.exceptions import InternalError


class ProgressBar(object):
    """A progress bar that automatically chooses the best display method."""

    def __init__(self, total, message=None, leave=True):
        env = Environment()
        if env.interactivity == Environment.Notebook:
            from tqdm import tqdm_notebook
            self._bar = tqdm_notebook(total=total, leave=leave)

            if message is not None:
                self._bar.set_description(message)
        else:
            from tqdm import tqdm
            self._bar = tqdm(total=total, leave=leave)

            if message is not None:
                self._bar.set_description(message)

        self.total = total

    @property
    def total(self):
        return self._bar.total

    @total.setter
    def total(self, total):
        self._bar.total = total

    def update(self, delta):
        self._bar.update(delta)

    def __enter__(self):
        self._bar.__enter__()

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._bar.__exit__(exc_type, exc_value, traceback)



"""Envelope calculation from a series of sampled functions."""

from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *
import numpy as np
from scipy.interpolate import interp1d
from typedargs.exceptions import ArgumentError
from .domain import combine_domains


def envelope(*arrays, **kwargs):
    """Return the minimum and maximum envelopes from a series of arrays.

    All passed in arguments should be 2D arrays.  The domain of all of the
    functions is found using combine_domains() and this is then divided into
    a fixed number of buckets using the bin_count keyword parameter (that
    must be passed as a keyword).

    For each bin in the combined domain, the minimum and maximum values in
    all of the passed arrays is then found and the resulting min/max envelopes
    are returned as a (bin_count, 3) array.  The first column is the bins that
    were used.  You can either return the left side, center or right side of
    the bins using the keyword bin_mark = "left" | "right" | "center".

    The default bin_mark is "center".

    Args:
        *arrays (ndarray): A list of numpy arrays that must be 2D and are used
             to compute a min/max envelope.
        bin_count (int): An optional keyword argument that indicates the number of
            bins to divde the domain into for envelope calculation.  Default: 100
        bin_spacing (str): An optional keyword argument of either 'linear' or 'log'
            to indicate whether the bins should be generated using.  Default: linear
        bin_mark (str): How to return the bins that were used to calculate the
            envelope.  Possible options are "left", "right" or "center" causing the
            first column of the output to contain the left edge, right edge or center
            of each bin respectively. Default: "center"
    """

    mark = kwargs.get('bin_mark', 'center')
    if mark not in ('left', 'right', 'center'):
        raise ArgumentError("Invalid bin_mark, must be left, right or center", bin_mark=mark)

    d_min, d_max = combine_domains(*arrays, type="union")

    num_bins = kwargs.get('bin_count', 100)

    bin_spacing = kwargs.get('bin_spacing', 'linear')
    if bin_spacing not in ('linear', 'log'):
        raise ArgumentError("Invalid bin spacing, must be linear or log", bin_spacing=bin_spacing)

    if bin_spacing == 'linear':
        bins = np.linspace(d_min, d_max, num_bins+1)
    else:
        bins = np.geomspace(d_min, d_max, num_bins+1)


    env_min = np.zeros_like(bins[1:])
    env_max = np.zeros_like(bins[1:])

    env_min[:] = np.nan
    env_max[:] = np.nan

    for arr in arrays:
        indices = np.digitize(arr[:,0], bins[1:], right=True)

        for arr_i, i in enumerate(indices):
            val = arr[arr_i, 1]

            if np.isnan(env_max[i]) or val > env_max[i]:
                env_max[i] = val
            if np.isnan(env_min[i]) or val < env_min[i]:
                env_min[i] = val

    centers = (bins[1:] + bins[:-1]) / 2.0

    # Fill in empty gaps with interpolation in case there weren't any values
    # in a given bucket.  We default to a simple linear interpolation
    nan_index = np.isnan(env_min)
    if np.any(nan_index):
        nonnan = np.logical_not(nan_index)
        min_interp = interp1d(centers[nonnan], env_min[nonnan], assume_sorted=True)
        env_min[nan_index] = min_interp(centers[nan_index])

    nan_index = np.isnan(env_max)
    if np.any(nan_index):
        nonnan = np.logical_not(nan_index)
        max_interp = interp1d(centers[nonnan], env_max[nonnan], assume_sorted=True)
        env_max[nan_index] = max_interp(centers[nan_index])


    out = np.ndarray([num_bins, 3])
    out[:, 1] = env_min
    out[:, 2] = env_max

    if mark == 'left':
        out[:, 0] = bins[:-1]
    elif mark == 'right':
        out[:, 0] = bins[1:]
    else:
        out[:, 0] = (bins[1:] + bins[:-1]) / 2.0



    return out



"""Basic configuration for interactive vs script environments."""
from .exceptions import MissingPackageError


class Environment(object):
    """Configuration class for better interactive computing experience."""

    Console = 0
    Notebook = 1
    Script = 2
    Unattended = 3

    _settings = {}

    def __init__(self):
        self.interactivity = Environment._settings.get('interactivity', Environment.Script)

    @classmethod
    def SetupNotebook(cls):
        """Configure iotile_analytics for a Jupyter Notebook.

        This changes the way input is requested and progress is
        displayed to be appropriate for an interactive notebook
        environment.  It also sets up matplotlib for inline
        graphics display.
        """

        try:
            import ipywidgets  # pylint:disable=W0612; We just want to test importability
        except ImportError:
            raise MissingPackageError("Missing required package ipywidgets", package="ipywidgets", suggestion="pip install ipywidgets")

        try:
            import tqdm  # pylint:disable=W0612; We just want to test importability
        except ImportError:
            raise MissingPackageError("Missing required package tqdm", package="tqdm", suggestion="pip install tqdm")

        try:
            from IPython import get_ipython

            ipython = get_ipython()
            ipython.magic("matplotlib inline")
        except ImportError:
            raise MissingPackageError("Missing required package IPython", package="IPython")

        cls._settings['interactivity'] = Environment.Notebook

    @classmethod
    def SetupScript(cls):
        """Configure iotile_analytics to run in an attended script.

        Attended scripts are those that are watched by a user in the
        command line but not run using jupyter-console.  They are run
        as normal python scripts.

        No graphical plotting is allowed since the shell doesn't allow
        it and progress bars are displayed in an ascii format.
        """

        try:
            import tqdm  # pylint:disable=W0612; We just want to test importability
        except ImportError:
            raise MissingPackageError("Missing required package tqdm", package="tqdm", suggestion="pip install tqdm")

        cls._settings['interactivity'] = Environment.Script



"""Tests for interacting with IOTile.cloud."""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

import os.path
from iotile_cloud.api.connection import Api


def config_path(filename):
    """Create a config path."""

    base = os.path.dirname(__file__)
    return os.path.join(base, 'data', filename)


def test_mock_cloud_login(mock_cloud):
    """Make sure the mock cloud is working."""

    domain, cloud = mock_cloud
    api = Api(domain=domain)

    res = api.login('test', 'test@arch-iot.com')
    assert res

    res = api.login('test2', 'test@arch-iot.com')
    assert not res

    res = api.login('test', 'test@arch-iot.com1')
    assert not res


def test_data_access(water_meter):
    """Make sure we can load and access data."""

    domain, _cloud = water_meter

    api = Api(domain=domain)
    api.login('test', 'test@arch-iot.com')

    data = api.device('d--0000-0000-0000-00d2').get()
    assert data['slug'] == 'd--0000-0000-0000-00d2'

    stream = api.stream('s--0000-0077--0000-0000-0000-00d2--5002').get()
    assert stream['slug'] == 's--0000-0077--0000-0000-0000-00d2--5002'

    proj = api.project('1c07fdd0-3fad-4549-bd56-5af2aca18d5b').get()
    assert proj['slug'] == 'p--0000-0077'



"""Determine the domain of a group of data series."""

from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

import pandas as pd
import numpy as np
from typedargs.exceptions import ArgumentError


def combine_domains(*args, **kwargs):
    """Determine the combined domain of a set of series.

    This function will return an interval that either contains
    the union of all domains passed in args, or the intersection,
    depending on whether type is "union" or "intersection".

    For example, if you pass in the following series:

    0, x1
    5, x2

    and

    -10, y1
    2, y2

    then the output with type="union" would be [-10, 5]
    with type="intersection", the output would be [2, 5]
    """

    domains = [find_domain(x) for x in args]

    ag_type = kwargs.get('type', 'union')
    if ag_type not in ('union', 'intersection'):
        raise ArgumentError("Unknown domain combination, must be union or intersection", type=ag_type)

    if ag_type == 'union':
        return (min([x[0] for x in domains]), max([x[1] for x in domains]))

    # We're asked for an intersection, keep a running intersection
    # and if there is not overlap among all of the intervals, throw an exception
    int_start, int_end = domains[0]

    for d_min, d_max in domains[1:]:
        if d_max < int_start or d_min > int_end:
            raise ArgumentError("Not all domains intersect in listed domains", running_intersection=(int_start, int_end), next_domain=(d_min, d_max))

        int_start = max(int_start, d_min)
        int_end = min(int_end, d_max)

    return (int_start, int_end)


def find_domain(input_data):
    """Find the min and max value of the domain of an input series.

    The input can either be a Pandas Series in which case the index
    is used if set, otherwise the first column is used.  If the input
    is a Numpy array, then the first column is used.

    Args:
        input_data (Series or ndarray): The input data that we want to
        find the domain of.

    Returns:
        (min, max): A tuple with the minimum and maximum values of the
            domain of the passed input_data.
    """

    axis = input_data

    if isinstance(input_data, pd.Series):
        if input_data.index is not None:
            axis = input_data.index.values
        else:
            axis = input_data.values

    if not isinstance(axis, np.ndarray):
        raise ArgumentError("Unknown input array in find_domain, expected numpy array", input_data=input_data, extracted_axis=axis)

    if len(axis.shape) == 2:
        axis = axis[:, 0]
    elif len(axis.shape) != 1:
        raise ArgumentError("Attempted to call find_domain on an array with more than 2 dimensions", input_data=input_data, shape=axis.shape)

    dmin = np.min(axis)
    dmax = np.max(axis)

    return (dmin, dmax)



"""High level entry point for iotile-analytics-core package."""

from .group import AnalysisGroup
from .session import CloudSession
from .interaction import ProgressBar

__all__ = ['AnalysisGroup', 'CloudSession', 'ProgressBar']



"""Simple session management system to save an iotile.cloud token."""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

import logging
import getpass
import math
import json

try:
    #python2
    from urllib2 import urlopen, Request
    from urllib import urlencode
except ImportError:
    #python3
    from urllib.request import urlopen, Request
    from urllib.parse import urlencode

from threading import Lock
from multiprocessing.pool import ThreadPool
from typedargs.exceptions import ArgumentError
from iotile_cloud.api.connection import Api, DOMAIN_NAME
from iotile_cloud.api.exceptions import HttpNotFoundError, HttpClientError, RestHttpBaseException
from .exceptions import CloudError, AuthenticationError
from .interaction import ProgressBar


logger = logging.getLogger(__name__)


class CloudSession(object):
    """A persistent session to IOTile.cloud.

    CloudSession internally caches results and information from iotile.cloud
    so that you don't have to login multiple times per script invocation.

    All caching is performed on a per domain basis so you can have multiple sessions
    going to different IOTile.cloud instances if needed.
    """

    MAX_CONCURRENCY = 10
    _login_cache = {}

    def __init__(self, user=None, password=None, domain=DOMAIN_NAME):
        self.pool = ThreadPool(CloudSession.MAX_CONCURRENCY)

        if domain not in CloudSession._login_cache:
            CloudSession._login_cache[domain] = {'requests': {}, 'request_lock': Lock()}

        cache = CloudSession._login_cache[domain]
        self.token = cache.get('token', None)
        self.token_type = cache.get('token_type', None)
        self.domain = domain

        self._api = Api(domain=domain)
        self._api.set_token(self.token, self.token_type)

        # Explictly log in if we're passed all of the right information
        # or our current token is invalid
        if self.token is not None and self.token_type is not None and user is None and password is None:
            return

        if password is None and self._check_token():
            return

        if user is None:
            user = raw_input("Please enter your IOTile.cloud email: ")

        if password is None:
            password = getpass.getpass("Please enter your IOTile.cloud password:")

        res = self._api.login(email=user, password=password)
        if not res:
            raise AuthenticationError("Could not login to IOTile.cloud", user=user, domain=domain)

        self.token = self._api.token
        self.token_type = self._api.token_type

        cache['token'] = self.token
        cache['token_type'] = self.token_type

    def _check_token(self):
        """Verify that we're able to log in to IOTile.cloud with our token."""

        if self.token is None:
            return False

        try:
            _info = self._api.account.get()

            return True
        except HttpClientError:
            return False

    def fetch_multiple(self, resources, per_call_kw=None, concurrency=MAX_CONCURRENCY, message=None, **kwargs):
        """Fetch multiple resources in parallel.

        Up to concurrency max calls are in flight at a given time.  The results
        are returned in the same order as resources is passed.

        Args:
            resources (RestResource[]): Should be created from an Api object.
            per_call_kw (dict[]): An optional set of keyword arguments that should be individual
                for each fetch call.  For example, you could pass a custom filter argument for
                each call.
            concurrency (int): The maximum number of parallel requests to send.
            message (str): Optional descriptive message that is printed with the progress bar
            **kwargs (str): Additional keyword arguments that are passed as part of
                the query string in the get request.

        Returns:
            dict[]: A list of the results for each resource.
        """

        if per_call_kw is None:
            per_call_kw = [{} for x in range(0, len(resources))]

        if len(per_call_kw) != len(resources):
            raise ArgumentError("You must pass the same number of per_call keyword arguments as resources (or none at all)", fetch_length=len(resources), kw_length=len(per_call_kw))

        try:
            with ProgressBar(total=len(resources), message=message, leave=False) as progbar:
                args = [(x, _merge_dicts(per_call_kw[i], kwargs), progbar) for i, x in enumerate(resources)]
                wrapped_results = self.pool.map(self._resource_fetcher, args)

                failed = [err for result, err in wrapped_results if err is not None]
                if len(failed) > 0:
                    raise failed[0]

                return [result for result, _err in wrapped_results]
        except RestHttpBaseException as err:
            raise self._translate_error(err, msg="Error fetching resources in parallel from IOTile.cloud")

    def fetch_all(self, resource, page_size=100, concurrency=MAX_CONCURRENCY, message=None, **kwargs):
        """Fetch and concatenate all pages of a given resource.

        The pages are fetched in parallel using up to concurrency requests
        at a time.

        If one page request fails, the fetch fails.  A single call is made to
        figure out how many results there are and then batched calls are made
        to download the pages in parallel.

        The RestResource is accessed using get with any keyword arguments passed
        to this function.

        **You cannot pass the page keyword argument to this function explicitly since
        that keyword is generated and used internally.**

        Args:
            resource (RestResource): Should be created from an Api object.
            page_size (int): The desired page size to use for fetches.
            concurrency (int): The maximum number of parallel requests to send.
            message (str): Optional descriptive message that is printed with the progress bar
            **kwargs (str): Additional keyword arguments that are passed as part of
                the query string in the get request.

        Returns:
            dict[]: A list of the results concatenated from all pages.
        """

        try:
            with ProgressBar(total=100, leave=False, message=message) as progbar:
                results = resource.get(page_size=page_size, **kwargs)
                total_count = results['count']
                results = results.get('results', [])

                if total_count <= page_size:
                    progbar.total = 1
                    progbar.update(1)
                    return results

                pages = int(math.ceil(total_count / float(page_size)))
                progbar.total = pages
                progbar.update(1)

                args = [(resource, x+1, page_size, kwargs, progbar) for x in range(1, pages)]

                wrapped_results = self.pool.map(self._url_fetcher, args)

                failed = [err for result, err in wrapped_results if err is not None]
                if len(failed) > 0:
                    raise failed[0]

                for result, _err in wrapped_results:
                    results.extend(result.get('results'))

                return results
        except RestHttpBaseException as err:
            raise self._translate_error(err, msg="Error fetching resource from IOTile.cloud", url=resource.url())

    def _resource_fetcher(self, args):
        resource, kwargs, progress = args

        try:
            url = resource.url()
            query_string = urlencode(kwargs)

            cache_key = "%s?%s" % (url, query_string)
            result = self._check_cache(cache_key)

            if result is None:
                headers = {b'Authorization': '{} {}'.format(self.token_type, self.token).encode('utf-8')}
                req = Request(cache_key, headers=headers)

                resp = urlopen(req)
                data = resp.read()
                result = json.loads(data.decode('utf-8'))
                #result = resource.get(**kwargs)
                self._cache_result(cache_key, result)

            progress.update(1)
            return result, None
        except Exception as err:  # pylint:disable=W0703; we do the exception processing in the calling function
            return None, err

    def _url_fetcher(self, args):
        """Method for fetching a url page in a worker thread."""

        resource, page, page_size, kwargs, progress = args

        try:
            url = resource.url()
            query = kwargs.copy()
            query['page_size'] = page_size
            query['page'] = page
            query_string = urlencode(query)

            cache_key = "%s?%s" % (url, query_string)
            result = self._check_cache(cache_key)

            if result is None:
                headers = {b'Authorization': '{} {}'.format(self.token_type, self.token).encode('utf-8')}
                req = Request(cache_key, headers=headers)

                resp = urlopen(req)
                data = resp.read()
                result = json.loads(data.decode('utf-8'))

                self._cache_result(cache_key, result)

            progress.update(1)
            return result, None
        except Exception as err:  # pylint:disable=W0703; we do the exception processing in the calling function
            return None, err

    def _cache_result(self, query, response):
        cache = self._login_cache[self.domain]

        with cache['request_lock']:
            cache['requests'][query] = response

    def _check_cache(self, key):
        cache = self._login_cache[self.domain]

        with cache['request_lock']:
            return cache['requests'].get(key, None)

    def get_api(self):
        """Return a logged in API object to IOTile.cloud.

        Returns:
            Api
        """

        return self._api

    @classmethod
    def _translate_error(cls, error, msg="Error interacting with iotile.cloud", **kwargs):
        """Translate a raw rest error into a user friendly cloud error."""

        #FIXME: Actually do the translation
        return CloudError(msg, raw_error=error, **kwargs)

    def find_device(self, slug=None, external_id=None):
        """Find an IOTile device by its slug or external_id.

        The device can be found either by passing its slug or numeric
        id directly or by passing a device specific external_id.

        External ids are typically used to link a virtual IOTile device
        with a separate physical object.  If you have a POD or Arch Node
        then you don't need to worry about external ids.

        **You must pass either a slug or an external_id but not both.**

        Args:
            slug (str): the slug of the device that we want.  This can either
                 be a short slug with leading zeros omitted or long slug. It
                 should start with 'd--' since it is a device slug.
            external_id (str): The external id of the device that we want to
                search for. This external_id must be unique among all devices
                so that it matches a single device.
        """

        if slug is not None and external_id is not None:
            raise ArgumentError("You must pass either a device slug or an external_id but not both", slug=slug, external_id=external_id)

        api = self.get_api()

        if external_id is not None:
            try:
                devices = api.device.get(external_id=external_id)
                results = devices.get('results', [])

                if len(results) < 0:
                    raise ArgumentError("No device matching external_id could be found", external_id=external_id, suggestion="Make sure the external id is correct")
                elif len(results) > 1:
                    slugs = [x['slug'] for x in results]
                    raise ArgumentError("More than one device with given external id found", external_id=external_id, found_devices=slugs)

                return results[0]
            except HttpClientError as err:
                raise self._translate_error(err)

        try:
            return api.device(slug).get()
        except HttpNotFoundError:
            raise ArgumentError("No device with the given slug could be found", slug=slug)
        except HttpClientError as err:
            raise self._translate_error(err, "Error fetching device information", slug=slug)

    def fetch_raw_event(self, event_or_id):
        """Fetch associated raw data for an event.

        Args:
            event_or_id (pd.Series or int): Either an event returned as a member of the response
                of fetch_events with the event_id key set or an id for an event as an integer.

        Returns:
            dict: The raw event object data fetched from iotile.cloud.
        """

        if hasattr(event_or_id, 'event_id'):
            event_id = event_or_id['event_id']
        else:
            event_id = event_or_id

        if not isinstance(event_id, int):
            try:
                event_id = int(event_id)
            except ValueError:
                raise ArgumentError("Invalid event passed to fetch_raw_event that could not be converted to an integer", event=event_or_id, event_id=event_id)

        try:
            api = self.get_api()
            return api.event(event_id).data.get()
        except HttpNotFoundError:
            raise ArgumentError("No event with the given id could be found", event_id=event_id)
        except HttpClientError as err:
            raise self._translate_error(err, "Error fetching event data", event_id=event_id)


def _merge_dicts(dict1, dict2):
    result = dict1.copy()
    result.update(dict2)
    return result



"""Tests of basic functionality."""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

def test_basic_import():
    """Make sure we can import."""

    import iotile_analytics.core



"""The AnalysisGroup is the main entry point for analyzing data.

It allows you to select data streams from IOTile.cloud and turn them
into Pandas time series for analysis.  You can create an AnalysisGroup
from a Project, a Device or a DataBlock.
"""

from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *
from future.utils import viewitems

import pandas as pd
from iotile_cloud.api.connection import DOMAIN_NAME
from iotile_cloud.api.exceptions import RestHttpBaseException
from iotile_cloud.stream.data import StreamData
from typedargs.exceptions import ArgumentError
from .exceptions import CloudError
from .session import CloudSession


class AnalysisGroup(object):
    """A top level entry point for selecting and analyzing data.

    Currently AnalysisGroups must be created by passing the id
    or slug of a project in iotile.cloud and then are populated with
    all of the data streams and variables in that project.

    If you specify 'device' for source_type, the cloud_id should be
    a slug for the device that you want to fetch streams from.

    If you specify 'archive' for the source type, the cloud_id should
    be a slug for the archive that you want to fetch streams from.

    if you specify 'project' for source type, the cloud_id should be
    the UUId of the project that you want to fetch streams from.

    Args:
        cloud_id (str): An IOTile.cloud id (usually a slug) that
            selects that data streams are included in this analysis
            project.
        source_type (str): The type of iotile.cloud object that we
            are linking this AnalysisProject to.  This defaults to
            an iotile.cloud project, but finding only streams from a
            single device is also supported.  You can currently
            specify 'project' ,'device' or 'archive'.
        domain (str): Optional iotile.cloud domain to connect to (defaults to
            https://iotile.cloud).
    """

    def __init__(self, cloud_id, source_type='project', domain=DOMAIN_NAME):
        stream_finders = {
            'project': self._find_project_streams,
            'device': self._find_device_streams,
            'archive': self._find_archive_streams
        }

        stream_finder = stream_finders.get(source_type, None)
        if stream_finder is None:
            raise ArgumentError("Invalid source type", source_type=source_type, supported_sources=stream_finders.keys(), suggestion="Try using one of the convenience functions for creating an AnalysisGroup like FromDevice or FromProject")

        session = CloudSession(domain=domain)
        self._session = session
        self._api = session.get_api()

        stream_list = stream_finder(cloud_id)
        self.streams = self._parse_stream_list(stream_list)
        self.stream_counts = self._count_streams([x['slug'] for x in stream_list])

        self._stream_table = [(slug.lower(), self._get_stream_name(stream).lower()) for slug, stream in viewitems(self.streams)]

    def _count_streams(self, slugs):
        resources = [self._api.stream(x).data for x in slugs]
        ts_results = self._session.fetch_multiple(resources, message='Counting Events in Streams', page_size=1)

        event_resources = [self._api.event for x in slugs]
        event_args = [{"filter": x} for x in slugs]

        event_results = self._session.fetch_multiple(event_resources, event_args, message='Counting Data in Streams', page_size=1)

        return {slug: {'points': ts['count'], 'events': event['count']} for slug, ts, event in zip(slugs, ts_results, event_results)}

    def stream_empty(self, slug):
        """Check if a stream is empty.

        Empty streams have neither event data nor time series data.

        Args:
            slug (str): The slug of the stream tha we want to check to
                see if it empty

        Returns:
            bool: True if the stream is empty, otherwise false.
        """

        counts = self.stream_counts[slug]
        return counts['points'] == 0 and counts['events'] == 0

    def print_streams(self, include_empty=False):
        """Print a table of the streams in this AnalysisGroup.

        By default this function does not print streams that have no data
        but you can override that by passing include_empty=True.

        Args:
            include_empty (bool): Also show streams that have no data.
        """

        print("{:40s} {:s}".format("Name", "Slug"))
        print("{:40s} {:s}".format("----", "----"))

        for slug, stream in self.streams.items():
            if self.stream_empty(slug) and not include_empty:
                continue

            name = self._get_stream_name(stream)

            if len(name) > 37:
                name = name[:37] + '...'

            print('{:40s} {:s}'.format(name, slug))

    @classmethod
    def _get_stream_name(cls, stream):
        name = stream.get('data_label', '')

        if name == '':
            name = stream.get('var_name') + "(from variable)"

        if name == '':
            name = "Unnamed"

        return name

    def find_stream(self, partial, include_empty=False):
        """Find a stream by a partial identifier.

        The identifier can either be a part of its slug or its name or the
        name of the variable that it comes from.  Whatever you pass must
        uniquely match exactly one stream in this analysis project or an
        exception will be thrown.

        Matching is done in a case insensitive fashion.

        By default, only streams that are not empty are searched.  You can
        override this by passing include_empty=True.  This behavior is the
        same as the print_streams() behavior, so searching for a stream based
        on what you find in print streams will work as expected.

        Args:
            partial (str): A partial slug or identifier to use to uniquely
                find one stream.
            include_empty (bool): Also include streams that have no data in
                the search.

        Returns:
            str: The full slug of the stream that we were looking for, suitable
                for passing to an IOTile.cloud api.

        Raises:
            ArgumentError: If the partial matches no streams or if it matches
                multiple streams.
        """

        if isinstance(partial, int):
            partial = "{:x}".format(partial)

        if not isinstance(partial, str):
            partial = str(partial)

        partial = partial.lower()

        found = [slug for slug, stream in self._stream_table if (partial in slug or partial in stream) and (include_empty or not self.stream_empty(slug))]
        if len(found) == 0:
            raise ArgumentError("No stream found matching given partial identifier", partial=partial)
        elif len(found) > 1:
            raise ArgumentError("Multiple matching streams found", partial=partial, slugs=found)

        return found[0]

    def fetch_stream(self, slug_or_name):
        """Fetch data from a stream by its slug or name.

        For example say you have the following stream in this analysis project:
        s--0000-0011--0000-0000-0000-0222--5001 (Temperature)

        You could fetch it by passing any of:
         - temp
         - 5001
         - Temperature
         - s--0000-0011--0000-0000-0000-0222--5001

        Args:
            slug_or_name (str): The stream that we want to fetch.  This
                can be a partial match to a full stream slug or name so long
                as it uniquely matches.  This is passed to find_stream so anything
                that find_stream accepts will be accepted here.
        """

        slug = self.find_stream(slug_or_name)
        data = self._get_stream_data(slug)

        dt_index = pd.to_datetime([x['timestamp'] for x in data])
        return pd.Series([x['output_value'] for x in data], index=dt_index)

    def fetch_events(self, slug_or_name):
        """Fetch event metadata from a stream by its slug or name.

        This function will return a Pandas DataFrame with all of the
        events in the stream.  It will return only the extra summary
        data stored with the event itself however.  If you need access to
        any raw data uploaded with the event that was not part of the
        summary, you need to use fetch_event for each event to get
        the raw data.

        slug_or_name is passed to find_stream to convert it into a slug so
        partial stream names are accepted.

        Args:
            slug_or_name (str): The stream that we want to fetch.  This
                can be a partial match to a full stream slug or name so long
                as it uniquely matches.  This is passed to find_stream so anything
                that find_stream accepts will be accepted here.

        Returns:
            DataFrame: All of the extra_data associated with the event as a
                Pandas DataFrame.
        """

        slug = self.find_stream(slug_or_name)

        try:
            data = self._get_event_data(slug)

            dt_index = pd.to_datetime([x['timestamp'] for x in data])
            extra_data = [x['extra_data'] for x in data]

            for i, event in enumerate(data):
                extra_data[i]['event_id'] = event['id']

            return pd.DataFrame(extra_data, index=dt_index)
        except RestHttpBaseException as exc:
            raise CloudError("Error fetching events from stream", exception=exc, response=exc.response.status_code)

    def fetch_raw_events(self, slug_or_name, subkey=None, postprocess=None):
        """Fetch multiple raw events by numeric id.

        Args:
            slug_or_name (str): The stream that we want to fetch.  This
                can be a partial match to a full stream slug or name so long
                as it uniquely matches.  This is passed to find_stream so anything
                that find_stream accepts will be accepted here.
            subkey (str): Only include a single key of the event.  This is usefuly for complex
                events where you want to just focus on a portion of them.
            postprocess (callable): (Optional) function to call on each raw event before
                adding it to the dataframe.

        Returns:
            pd.DataFrame: The raw event object data fetched from iotile.cloud.
        """

        events = self.fetch_events(slug_or_name)
        event_ids = events['event_id'].values

        resources = [self._api.event(x).data for x in event_ids]
        data = self._session.fetch_multiple(resources)

        if postprocess is None:
            postprocess = lambda x: x

        if subkey is not None:
            data = [postprocess(x[subkey]) for x in data]
        else:
            data = [postprocess(x) for x in data]

        return pd.DataFrame(data, index=events.index)

    def fetch_raw_event(self, event_or_id, subkey=None):
        """Fetch associated raw data for an event.

        Args:
            event_or_id (pd.Series or int): Either an event returned as a member of the response
                of fetch_events with the event_id key set or an id for an event as an integer.
            subkey (str): Only include a single key of the event.  This is usefuly for complex
                events where you want to just focus on a portion of them.

        Returns:
            pd.DataFrame: The raw event object data fetched from iotile.cloud.
        """

        event = self._session.fetch_raw_event(event_or_id)

        if subkey is not None:
            if subkey not in event:
                raise ArgumentError("Desired key was not found in event", subkey=subkey, keys=event.keys())

            event = event[subkey]

        return pd.DataFrame.from_dict(event)

    def _find_project_streams(self, project_id):
        """Find all streams in a project by its uuid."""

        try:
            results = self._api.stream.get(project=project_id)
            return results['results']
        except RestHttpBaseException as exc:
            raise CloudError("Error calling method on iotile.cloud", exception=exc, response=exc.response.status_code)

    def _find_device_streams(self, device_slug):
        """Find all streams for a device by its slug."""

        try:
            results = self._api.stream.get(device=device_slug)
            return results['results']
        except RestHttpBaseException as exc:
            raise CloudError("Error calling method on iotile.cloud", exception=exc, response=exc.response.status_code)

    def _find_archive_streams(self, archive_slug):
        """Find all streams for an archive by its slug."""

        try:
            results = self._api.stream.get(block=archive_slug, all=1)
            return results['results']
        except RestHttpBaseException as exc:
            raise CloudError("Error calling method on iotile.cloud", exception=exc, response=exc.response.status_code)

    @classmethod
    def _parse_stream_list(cls, stream_list):
        return {stream['slug']: stream for stream in stream_list}

    def _get_stream_data(self, slug, start=None, end=None):
        resource = self._api.stream(slug).data
        return self._session.fetch_all(resource, page_size=1000, message="Downloading Stream Data")

    def _get_event_data(self, slug):
        resource = self._api.event
        return self._session.fetch_all(resource, page_size=1000, message="Downloading Events", filter=slug)

    @classmethod
    def FromDevice(cls, slug=None, external_id=None, domain=DOMAIN_NAME):
        """Create a new AnalysisGroup from a single device.

        The device can be found either by passing its slug or numeric
        id directly or by passing a device specific external_id.

        External ids are typically used to link a virtual IOTile device
        with a separate physical object.  If you have a POD or Arch Node
        then you don't need to worry about external ids.

        **You must pass either a slug or an external_id but not both.**

        Args:
            slug (str): the slug of the device that we want.  This can either
                 be a short slug with leading zeros omitted or long slug. It
                 should start with 'd--' since it is a device slug.
            external_id (str): The external id of the device that we want to
                search for. This external_id must be unique among all devices
                so that it matches a single device.
            domain (str): Optional iotile.cloud domain to connect to (defaults to
                https://iotile.cloud).
        """

        session = CloudSession(domain=domain)
        device = session.find_device(slug=slug, external_id=external_id)

        return AnalysisGroup(device['slug'], source_type="device", domain=domain)

    @classmethod
    def FromArchive(cls, slug=None, domain=DOMAIN_NAME):
        """Create a new AnalysisGroup from a single archive.

        The archive is found by passing its archive slug:
        b--XXXX-XXXX-XXXX-XXXX

        Args:
            slug (str): the slug of the archive that we want.  This can either
                 be a short slug with leading zeros omitted or long slug. It
                 should start with 'b--' since it is an archive slug.
            domain (str): Optional iotile.cloud domain to connect to (defaults to
                https://iotile.cloud).
        """

        return AnalysisGroup(slug, source_type="archive")



"""Tests to make sure domain union/intersection and finding works."""

from iotile_analytics.core.utilities import find_domain, combine_domains
from typedargs.exceptions import ArgumentError
import pandas as pd
import numpy as np
import pytest


def test_find_domain():
    """Make sure we can find domains."""

    arr1 = np.linspace(0, 10, 10)

    arr2 = np.ndarray([10, 2])
    arr2[:, 0] = arr1
    arr2[:, 1] = 1

    arr3 = pd.Series(np.zeros_like(arr1), index=arr1)

    dom1 = find_domain(arr1)
    dom2 = find_domain(arr2)
    dom3 = find_domain(arr3)

    assert dom1 == (0, 10)
    assert dom2 == dom1
    assert dom3 == dom2


def test_combine_domains():
    """Make sure out domain combination works."""

    arr1 = np.linspace(0, 10)
    arr2 = np.linspace(5, 15)
    arr3 = np.linspace(-5, 5)
    arr4 = np.linspace(-10, -5)

    assert combine_domains(arr1, arr2) == (0, 15)
    assert combine_domains(arr1, arr2, type="intersection") == (5, 10)

    with pytest.raises(ArgumentError):
        combine_domains(arr1, arr4, type="intersection")

    assert combine_domains(arr1, arr2, arr3, arr4) == (-10, 15)



"""Interactivity primitives that work across environments.

In particular, we need a way to ask a user a question if needed
and show the user progress.  However, the best way to do this
varies depending on whether the user's sesion is running in a
shell script, a jupyter notebook, a jupyter console or on
an unattended background worker.
"""

from .progress import ProgressBar

__all__ = ['ProgressBar']



"""Custom exceptions for iotile analytics."""

from typedargs.exceptions import KeyValueException


class CloudError(KeyValueException):
    """A generic error interacting with IOTile.cloud."""

    pass


class AuthenticationError(CloudError):
    """An error logging into iotile.cloud."""

    pass


class MissingPackageError(KeyValueException):
    """The operation you requested requires a packgae to be installed.

    The suggestion keyword argument should be the name of the package.
    """

    pass



"""Setup file for iotile_analytics."""



from setuptools import setup, find_packages
from version import version

setup(
    name="iotile-analytics-core",
    packages=find_packages(exclude=("test",)),
    version=version,
    license="LGPLv3",
    install_requires=[
        "future>=0.16.0",
        "numpy>=1.13.1",
        "pandas>=0.20.3",
        "scipy>=1.0.0",
        "typedargs>=0.10.0",
        "iotile_cloud>=0.7.0",
        "tqdm>=4.19.4"
    ],
    description="A data science bridge for iotile.cloud",
    author="Arch",
    author_email="info@arch-iot.com",
    url="https://github.com/iotile/typedargs",
    keywords=[""],
    classifiers=[
        "Programming Language :: Python",
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules"
        ],
    long_description="""\
IOTile Analytics Core
----------------------

IOTile Analytics Core provides classes for pulling data from iotile.cloud into
a standard data science environment.  In particular it natively maps data into
numpy and pandas as appropriate and provides functions to select data.  It
works well with a Jupyter Notebook based analysis environment, through the
functionality can be used in any python environment.

The IOTile Analytics package is broken into multiple parts of which
iotile-analytics-core is one.  You should install the additional parts that you need for
your use cases.

Installation:

```
pip install iotile_analytics
```
"""
)
